import 'package:dynatrace_flutter_plugin/dynatrace_flutter_plugin.dart';
import 'package:dynatrace_flutter_plugin/src/agent/action/action.dart';
import 'package:dynatrace_flutter_plugin/src/agent/agent_impl.dart';
import 'package:dynatrace_flutter_plugin/src/agent/action/counter.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:mockito/mockito.dart';

import '../../util/mock_method_channel.dart';
import '../../widget.dart';

void main() {
  late MockMethodChannel methodChannel;
  late Dynatrace dynatrace;

  setUp(() async {
    methodChannel = MockMethodChannel();

    Map<String, bool> configuration = {};
    configuration["autoStart"] = true;
    configuration["crashReporting"] = true;
    configuration["webRequest"] = true;
    configuration["logLevel"] = true;

    when(methodChannel.invokeMapMethod("getAutoStartConfiguration"))
        .thenAnswer((_) => Future<Map<String, bool>>.value(configuration));

    dynatrace = DynatraceImpl.private(methodChannel);
    await dynatrace.start(const TestMainWidget());
  });

  test('Action counting', () {
    DynatraceRootAction rootAction = dynatrace.enterAction("RootAction");

    rootAction.enterAction("Test");
    verify(methodChannel.invokeMethod(
            any, {"key": 1, "parent": 0, "name": "Test", "platform": null}))
        .called(1);

    rootAction.enterAction("AnotherAction");
    verify(methodChannel.invokeMethod(any, {
      "key": 2,
      "parent": 0,
      "name": "AnotherAction",
      "platform": null
    })).called(1);
  });

  test('Create Action when closed', () {
    DynatraceRootAction rootAction =
        DynatraceRootActionImpl.private(0, methodChannel);

    rootAction.enterAction("Test");
    verify(methodChannel.invokeMethod(any, any)).called(1);

    rootAction.leaveAction();
    DynatraceAction action = rootAction.enterAction("AnotherAction");
    action.reportEvent("eventName");
    verifyNever(methodChannel.invokeMethod(any));
  });

  test('Null Action Creation', () {
    DynatraceRootAction rootAction =
        DynatraceRootActionImpl.private(0, methodChannel);

    DynatraceAction nullAction = rootAction.enterAction("");

    nullAction
      ..reportDoubleValue("valueName", 1.0)
      ..reportError("errorName", -1)
      ..reportEvent("eventName")
      ..reportIntValue("valueName", 1)
      ..reportStringValue("valueName", "value")
      ..leaveAction();

    nullAction = rootAction.enterAction(null);

    nullAction
      ..reportDoubleValue("valueName", 1.0)
      ..reportError("errorName", -1)
      ..reportEvent("eventName")
      ..reportIntValue("valueName", 1)
      ..reportStringValue("valueName", "value")
      ..leaveAction();

    verifyNever(methodChannel.invokeMethod(any));
  });

  test('Null Root Action Creation', () {
    DynatraceRootAction nullAction = dynatrace.enterAction("");
    nullAction
      ..reportDoubleValue("valueName", 1.0)
      ..reportError("errorName", -1)
      ..reportEvent("eventName")
      ..reportIntValue("valueName", 1)
      ..reportStringValue("valueName", "value")
      ..leaveAction();

    nullAction = dynatrace.enterAction("");
    nullAction
      ..reportDoubleValue("valueName", 1.0)
      ..reportError("errorName", -1)
      ..reportEvent("eventName")
      ..reportIntValue("valueName", 1)
      ..reportStringValue("valueName", "value")
      ..leaveAction();

    DynatraceAction nullNullAction = nullAction.enterAction("");
    nullNullAction
      ..reportDoubleValue("valueName", 1.0)
      ..reportError("errorName", -1)
      ..reportEvent("eventName")
      ..reportIntValue("valueName", 1)
      ..reportStringValue("valueName", "value")
      ..leaveAction();

    verifyNever(methodChannel.invokeMethod(any));
  });

  group('Action Reporting', () {
    test('reportError', () {
      DynatraceRootAction rootAction =
          DynatraceRootActionImpl.private(0, methodChannel);

      rootAction
        ..reportError("", -1)
        ..reportError(null, -1)
        ..reportError("Test", null);

      verifyNever(methodChannel.invokeMethod(any));

      rootAction.reportError("ErrorName", -1);
      verify(methodChannel.invokeMethod(any, {
        "key": 0,
        "errorName": "ErrorName",
        "errorCode": -1,
        "platform": null
      })).called(1);

      rootAction
        ..reportError("ErrorName", -1, platform: Platform.iOS)
        ..leaveAction()
        ..reportError("ErrorName", -1, platform: Platform.iOS);

      verify(methodChannel.invokeMethod(any, {
        "key": 0,
        "errorName": "ErrorName",
        "errorCode": -1,
        "platform": 1
      })).called(1);
    });

    test('reportEvent', () {
      DynatraceRootAction rootAction =
          DynatraceRootActionImpl.private(0, methodChannel);

      rootAction
        ..reportEvent(null)
        ..reportEvent("");

      verifyNever(methodChannel.invokeMethod(any));

      rootAction.reportEvent("EventName");
      verify(methodChannel.invokeMethod(
          any, {"key": 0, "name": "EventName", "platform": null})).called(1);

      rootAction
        ..reportEvent("EventName", platform: Platform.Android)
        ..leaveAction()
        ..reportEvent("EventName", platform: Platform.Android);

      verify(methodChannel.invokeMethod(
          any, {"key": 0, "name": "EventName", "platform": 0})).called(1);
    });

    test('reportStringValue', () {
      DynatraceRootAction rootAction =
          DynatraceRootActionImpl.private(0, methodChannel);

      rootAction
        ..reportStringValue(null, null)
        ..reportStringValue("", "")
        ..reportStringValue("Test", "")
        ..reportStringValue("", "Test")
        ..reportStringValue("Test", null)
        ..reportStringValue(null, "Test");

      verifyNever(methodChannel.invokeMethod(any));

      rootAction.reportStringValue("ValueName", "Value");
      verify(methodChannel.invokeMethod(any, {
        "key": 0,
        "name": "ValueName",
        "value": "Value",
        "platform": null
      })).called(1);

      rootAction
        ..reportStringValue("ValueName", "Value", platform: Platform.Android)
        ..leaveAction()
        ..reportStringValue("ValueName", "Value", platform: Platform.Android);

      verify(methodChannel.invokeMethod(any, {
        "key": 0,
        "name": "ValueName",
        "value": "Value",
        "platform": 0
      })).called(1);
    });

    test('reportIntValue', () {
      DynatraceRootAction rootAction =
          DynatraceRootActionImpl.private(0, methodChannel);

      rootAction
        ..reportIntValue(null, null)
        ..reportIntValue("Test", null)
        ..reportIntValue("", -1)
        ..reportIntValue(null, -1);

      verifyNever(methodChannel.invokeMethod(any));

      rootAction.reportIntValue("ValueName", 12);
      verify(methodChannel.invokeMethod(any, {
        "key": 0,
        "name": "ValueName",
        "value": 12,
        "platform": null
      })).called(1);

      rootAction
        ..reportIntValue("ValueName", 12, platform: Platform.iOS)
        ..leaveAction()
        ..reportIntValue("ValueName", 12, platform: Platform.iOS);

      verify(methodChannel.invokeMethod(
              any, {"key": 0, "name": "ValueName", "value": 12, "platform": 1}))
          .called(1);
    });

    test('reportDoubleValue', () {
      DynatraceRootAction rootAction =
          DynatraceRootActionImpl.private(0, methodChannel);

      rootAction
        ..reportDoubleValue(null, null)
        ..reportDoubleValue("Test", null)
        ..reportDoubleValue("", -1.1)
        ..reportDoubleValue(null, -1.1);

      verifyNever(methodChannel.invokeMethod(any));

      rootAction.reportDoubleValue("ValueName", 12.1);
      verify(methodChannel.invokeMethod(any, {
        "key": 0,
        "name": "ValueName",
        "value": 12.1,
        "platform": null
      })).called(1);

      rootAction
        ..reportDoubleValue("ValueName", 12.1, platform: Platform.Android)
        ..leaveAction()
        ..reportDoubleValue("ValueName", 12.1, platform: Platform.Android);

      verify(methodChannel.invokeMethod(any, {
        "key": 0,
        "name": "ValueName",
        "value": 12.1,
        "platform": 0
      })).called(1);
    });
  });
  group('Action Reporting from static', () {
    test('reportError', () {
      DynatraceRootAction rootAction = dynatrace.enterAction("testError");
      verify(methodChannel.invokeMethod(any, {
        "name": "testError",
        "key": ActionCounter().getCount() - 1,
        "platform": null
      })).called(1);

      dynatrace
        ..reportErrorInAction("", -1)
        ..reportErrorInAction(null, -1)
        ..reportErrorInAction("Test", null);

      verifyNever(methodChannel.invokeMethod(any));

      dynatrace.reportErrorInAction("ErrorName", -1);

      verify(methodChannel.invokeMethod(any, {
        "key": ActionCounter().getCount() - 1,
        "errorName": "ErrorName",
        "errorCode": -1,
        "platform": null
      })).called(1);

      // dynatrace.reportErrorInAction("ErrorName", -1, platform: Platform.iOS);
      rootAction.leaveAction();
    });

    test('reportEvent', () {
      DynatraceRootAction rootAction = dynatrace.enterAction("testEvent");
      verify(methodChannel.invokeMethod(any, {
        "name": "testEvent",
        "key": ActionCounter().getCount() - 1,
        "platform": null
      })).called(1);

      dynatrace
        ..reportEvent(null)
        ..reportEvent("");

      verifyNever(methodChannel.invokeMethod(any));

      dynatrace.reportEvent("EventName");
      verify(methodChannel.invokeMethod(any, {
        "key": ActionCounter().getCount() - 1,
        "name": "EventName",
        "platform": null
      })).called(1);

      dynatrace.reportEvent("EventName", platform: Platform.Android);
      rootAction.leaveAction();
    });

    test('reportStringValue', () {
      DynatraceRootAction rootAction = dynatrace.enterAction("testStringValue");
      verify(methodChannel.invokeMethod(any, {
        "name": "testStringValue",
        "key": ActionCounter().getCount() - 1,
        "platform": null
      })).called(1);

      dynatrace
        ..reportStringValue(null, null)
        ..reportStringValue("", "")
        ..reportStringValue("Test", "")
        ..reportStringValue("", "Test")
        ..reportStringValue("Test", null)
        ..reportStringValue(null, "Test");

      verifyNever(methodChannel.invokeMethod(any));

      dynatrace.reportStringValue("ValueName", "Value");
      verify(methodChannel.invokeMethod(any, {
        "key": ActionCounter().getCount() - 1,
        "name": "ValueName",
        "value": "Value",
        "platform": null
      })).called(1);

      dynatrace.reportStringValue("ValueName", "Value",
          platform: Platform.Android);
      rootAction.leaveAction();
    });

    test('reportIntValue', () {
      DynatraceRootAction rootAction = dynatrace.enterAction("testIntValue");
      verify(methodChannel.invokeMethod(any, {
        "name": "testIntValue",
        "key": ActionCounter().getCount() - 1,
        "platform": null
      })).called(1);

      dynatrace
        ..reportIntValue(null, null)
        ..reportIntValue("Test", null)
        ..reportIntValue("", -1)
        ..reportIntValue(null, -1);

      verifyNever(methodChannel.invokeMethod(any));

      dynatrace.reportIntValue("ValueName", 12);
      verify(methodChannel.invokeMethod(any, {
        "key": ActionCounter().getCount() - 1,
        "name": "ValueName",
        "value": 12,
        "platform": null
      })).called(1);

      dynatrace.reportIntValue("ValueName", 12, platform: Platform.iOS);

      rootAction.leaveAction();
    });

    test('reportDoubleValue', () {
      DynatraceRootAction rootAction = dynatrace.enterAction("testDoubleValue");
      verify(methodChannel.invokeMethod(any, {
        "name": "testDoubleValue",
        "key": ActionCounter().getCount() - 1,
        "platform": null
      })).called(1);

      dynatrace
        ..reportDoubleValue(null, null)
        ..reportDoubleValue("Test", null)
        ..reportDoubleValue("", -1.1)
        ..reportDoubleValue(null, -1.1);

      verifyNever(methodChannel.invokeMethod(any));

      dynatrace.reportDoubleValue("ValueName", 12.1);
      verify(methodChannel.invokeMethod(any, {
        "key": ActionCounter().getCount() - 1,
        "name": "ValueName",
        "value": 12.1,
        "platform": null
      })).called(1);

      dynatrace.reportDoubleValue("ValueName", 12.1,
          platform: Platform.Android);

      rootAction.leaveAction();
    });

    /// Cancel Action

    test('Report something onto canceled Action', () {
      DynatraceRootAction rootAction =
          DynatraceRootActionImpl.private(0, methodChannel);

      rootAction.cancel();
      verify(methodChannel.invokeMethod("cancelAction", any)).called(1);

      rootAction.reportEvent("Test");

      verifyNever(methodChannel.invokeMethod("reportEvent"));
    });

    test('Cancel Action when already closed', () {
      DynatraceRootAction rootAction =
          DynatraceRootActionImpl.private(0, methodChannel);

      rootAction.leaveAction();
      verify(methodChannel.invokeMethod("leaveAction", any)).called(1);

      rootAction.cancel();
      verifyNever(methodChannel.invokeMethod("cancelAction"));
    });

    test('Cancel Action', () {
      DynatraceRootAction rootAction =
          DynatraceRootActionImpl.private(0, methodChannel);

      rootAction.cancel();
      verify(methodChannel.invokeMethod("cancelAction", any)).called(1);
    });
  });
}
