import 'package:dynatrace_flutter_plugin/dynatrace_flutter_plugin.dart';
import 'package:dynatrace_flutter_plugin/src/agent/agent_impl.dart';
import 'package:dynatrace_flutter_plugin/src/agent/util/logger.dart';
import 'package:flutter/material.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:mockito/mockito.dart';

import '../util/mock_method_channel.dart';
import '../widget.dart';

class MockLogger extends Mock implements Logger {}

void main() {
  late MockMethodChannel methodChannel;
  late DynatraceImpl dynatrace;

  setUp(() async {
    methodChannel = MockMethodChannel();

    Map<String, bool> configuration = {};
    configuration["autoStart"] = true;
    configuration["crashReporting"] = true;
    configuration["webRequest"] = true;
    configuration["logLevel"] = true;

    when(methodChannel.invokeMapMethod("getAutoStartConfiguration"))
        .thenAnswer((_) => Future<Map<String, bool>>.value(configuration));

    dynatrace = DynatraceImpl.private(methodChannel);
  });

  group('Agent Reporting', () {
    test('reportError', () async {
      await dynatrace.start(const TestMainWidget(),
          configuration: const Configuration());

      dynatrace
        ..reportError("", -1)
        ..reportError(null, -1)
        ..reportError("Test", null);

      verifyNever(methodChannel.invokeMethod(any));

      dynatrace.reportError("ErrorName", -1);
      verify(methodChannel.invokeMethod(any, {
        "errorName": "ErrorName",
        "errorCode": -1,
        "platform": null
      })).called(1);

      dynatrace.reportError("ErrorName", -1, platform: Platform.iOS);
      verify(methodChannel.invokeMethod(
              any, {"errorName": "ErrorName", "errorCode": -1, "platform": 1}))
          .called(1);
    });

    test('reportCrash', () async {
      await dynatrace.start(const TestMainWidget(),
          configuration: const Configuration());

      dynatrace
        ..reportCrash("", "reason", "stacktrace")
        ..reportCrash(null, "reason", "stacktrace")
        ..reportCrash(" ", "reason", "stacktrace");

      verifyNever(methodChannel.invokeMethod(any));

      dynatrace = DynatraceImpl.private(methodChannel);
      await dynatrace.start(const TestMainWidget(),
          configuration: const Configuration());
      dynatrace.reportCrash("errorName", "reason", "stacktrace");

      verify(methodChannel.invokeMethod("reportCrash", {
        "errorValue": "errorName",
        "reason": "reason",
        "stacktrace": "stacktrace",
        "platform": null
      })).called(1);
    });

    group('sendEvent', () {
      test('should not be possible to send an event when not started', () {
        dynatrace.sendEvent('customName', {'custom': 'field'});
        verifyNever(methodChannel.invokeMethod(any));
      });

      test('should not be possible to send an empty name', () async {
        await dynatrace.start(const TestMainWidget(),
            configuration: const Configuration());

        dynatrace.sendEvent('', null);
        verifyNever(methodChannel.invokeMethod(any));
      });

      test('should be possible to send an event', () async {
        await dynatrace.start(const TestMainWidget(),
            configuration: const Configuration());

        dynatrace.sendEvent('customName', {'custom': 'field'});
        verify(methodChannel.invokeMethod("sendEvent", {
          "name": "customName",
          "attributes": {/*'dt.agent.flavor': 'flutter', */ 'custom': 'field'},
          "platform": null
        })).called(1);
      });

      /*test('should be possible to override dt.agent.flavor', () async {
        await dynatrace.start(const TestMainWidget(),
            configuration: const Configuration());

        dynatrace.sendEvent('customName', {/*'dt.agent.flavor': 'flutter', */});
        verify(methodChannel.invokeMethod("sendEvent", {
          "name": "customName",
          "attributes": {/*'dt.agent.flavor': 'flutter', */},
          "platform": null
        })).called(1);
      });*/
    });

    group('sendBizEvent', () {
      test('should not be possible to send an event when not started', () {
        dynatrace.sendBizEvent('customName', {'custom': 'field'});
        verifyNever(methodChannel.invokeMethod(any));
      });

      test('should not be possible to send an empty name', () async {
        await dynatrace.start(const TestMainWidget(),
            configuration: const Configuration());

        dynatrace.sendBizEvent('', null);
        verifyNever(methodChannel.invokeMethod(any));
      });

      test('should be possible to send an event', () async {
        await dynatrace.start(const TestMainWidget(),
            configuration: const Configuration());

        dynatrace.sendBizEvent('customType', {'custom': 'field'});
        verify(methodChannel.invokeMethod("sendBizEvent", {
          "type": "customType",
          "attributes": {/*'dt.agent.flavor': 'flutter', */ 'custom': 'field'},
          "platform": null
        })).called(1);
      });

      /*test('should be possible to override dt.agent.flavor', () async {
        await dynatrace.start(const TestMainWidget(),
            configuration: const Configuration());

        dynatrace.sendBizEvent('customType', {'dt.agent.flavor': 'field'});
        verify(methodChannel.invokeMethod("sendBizEvent", {
          "type": "customType",
          "attributes": {'dt.agent.flavor': 'field'},
          "platform": null
        })).called(1);
      });*/
    });

    test('identifyUser', () async {
      await dynatrace.start(const TestMainWidget(),
          configuration: const Configuration());

      dynatrace
        ..identifyUser("")
        ..identifyUser(null);

      verifyNever(methodChannel.invokeMethod(any));

      dynatrace.identifyUser("User");
      verify(methodChannel
          .invokeMethod(any, {"user": "User", "platform": null})).called(1);

      dynatrace.identifyUser("User", platform: Platform.Android);
      verify(methodChannel.invokeMethod(any, {"user": "User", "platform": 0}))
          .called(1);
    });
  });

  test('Auto Startup', () async {
    // Check Startup with missing config
    await dynatrace.start(const TestMainWidget());
    verifyNever(methodChannel.invokeMethod("start"));
  });

  test('Manual Startup', () async {
    Map<String, bool> configuration = {};
    configuration["autoStart"] = false;
    configuration["crashReporting"] = true;
    configuration["webRequest"] = true;
    configuration["logLevel"] = true;

    when(methodChannel.invokeMapMethod("getAutoStartConfiguration"))
        .thenAnswer((_) => Future<Map<String, bool>>.value(configuration));

    // Check Startup with missing config
    dynatrace = DynatraceImpl.private(methodChannel);
    await dynatrace.start(const TestMainWidget(),
        configuration: const Configuration(beaconUrl: "url"));
    verifyNever(methodChannel.invokeMethod("start"));

    // Check if startup is fired if configuration is correct
    dynatrace = DynatraceImpl.private(methodChannel);
    await dynatrace.start(const TestMainWidget(),
        configuration:
            const Configuration(applicationId: "appId", beaconUrl: "url"));
    verify(methodChannel.invokeMethod(
        "start", {"applicationId": "appId", "beaconUrl": "url"})).called(1);
  });

  test('Manual Startup with startWithoutWidget()', () async {
    Map<String, bool> configuration = {};
    configuration["autoStart"] = false;
    configuration["crashReporting"] = true;
    configuration["webRequest"] = true;
    configuration["logLevel"] = true;

    when(methodChannel.invokeMapMethod("getAutoStartConfiguration"))
        .thenAnswer((_) => Future<Map<String, bool>>.value(configuration));

    // Check Startup with missing config
    dynatrace = DynatraceImpl.private(methodChannel);
    await dynatrace.startWithoutWidget(
        configuration: const Configuration(beaconUrl: "url"));
    runApp(const TestMainWidget());
    verifyNever(methodChannel.invokeMethod("start"));

    // Check if startup is fired if configuration is correct
    dynatrace = DynatraceImpl.private(methodChannel);
    await dynatrace.startWithoutWidget(
        configuration:
            const Configuration(applicationId: "appId", beaconUrl: "url"));
    runApp(const TestMainWidget());
    verify(methodChannel.invokeMethod(
        "start", {"applicationId": "appId", "beaconUrl": "url"})).called(1);
  });

  /**
   * reportZoneStacktrace
   */

  group('reportZoneStacktrace', () {

    MockLogger mockLogger = MockLogger();
    Logger().instance = mockLogger;

    setUp(() {
      reset(mockLogger);
    });

    test('when Flutter Plugin is not enabled', () {
      dynatrace.reportZoneStacktrace({}, StackTrace.empty);
      verify(mockLogger.i("Dynatrace reportZoneStacktrace(error, stacktrace): Dynatrace().start() method was not called!", logType: LogType.Warning)).called(1);
      verifyNever(methodChannel.invokeMethod(any));
    });

    test('when Message in Error Object is null', () async {
      await dynatrace.start(const TestMainWidget(),
          configuration: const Configuration());
      
      dynatrace.reportZoneStacktrace("Hello", StackTrace.empty);
      verify(mockLogger.d("Dynatrace reportZoneStacktrace(String: , )", logType: LogType.Info)).called(1);
      verify(methodChannel.invokeMethod("reportErrorStacktrace", any)).called(1);
    });

    test('when Error object is null', () async {
      await dynatrace.start(const TestMainWidget(),
          configuration: const Configuration());
      
      dynatrace.reportZoneStacktrace(null, StackTrace.empty);
      verify(mockLogger.d("Dynatrace reportZoneStacktrace(error, stacktrace): Error will not be reported because something is missing! \nerrorName: Null:  \nstacktrace: ", logType: LogType.Error)).called(1);
      verifyNever(methodChannel.invokeMethod("reportErrorStacktrace", any));
    });

    test('when Error object is correctly set', () async {
      await dynatrace.start(const TestMainWidget(),
            configuration: const Configuration());
        
      dynatrace.reportZoneStacktrace(Exception("FormatException"), StackTrace.empty);
      verify(mockLogger.d("Dynatrace reportZoneStacktrace(_Exception: FormatException, )", logType: LogType.Info)).called(1);
      verify(methodChannel.invokeMethod("reportErrorStacktrace", any)).called(1);
    });
  });
}
