import 'dart:async';

import 'package:dynatrace_flutter_plugin/dynatrace_flutter_plugin.dart';
import 'package:dynatrace_flutter_plugin/src/agent/agent_impl.dart';
import 'package:flutter/material.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:mockito/mockito.dart';

import '../util/mock_method_channel.dart';
import '../widget.dart';

dynamic error, context, stacktrace;

class MockMethodClass extends Mock {
  void mockFunction(Widget? _topLevelWidget);
}

void fakeRunApp(Widget _topLevelWidget) {
  // ignore: avoid_print
  print("Called Fake Run App Function");
  throw const FormatException();
}

void main() {
  late MockMethodChannel methodChannel;

  setUp(() {
    methodChannel = MockMethodChannel();

    Map<String, bool> configuration = {};
    configuration["autoStart"] = true;
    configuration["crashReporting"] = true;
    configuration["webRequest"] = true;
    configuration["logLevel"] = true;

    when(methodChannel.invokeMapMethod("getAutoStartConfiguration"))
        .thenAnswer((_) => Future<Map<String, bool>>.value(configuration));
  });

  test('Run App when Crash Reporting deactivated', () async {
    MockMethodClass method = MockMethodClass();
    Dynatrace dynatrace = DynatraceImpl.private(methodChannel,
        startUpFunction: method.mockFunction);

    await dynatrace.start(const TestMainWidget());

    verify(method.mockFunction(any)).called(1);
  });

  testWidgets('Test error reporting', (WidgetTester tester) async {
    Dynatrace dynatrace = DynatraceImpl.private(methodChannel);

    await dynatrace.start(const TestMainWidget());

    Widget main = MaterialApp(
      home: ElevatedButton(
        child: const Text('Test Error'),
        onPressed: () {
          throw const FormatException("Invalid Double");
        },
      ),
    );

    await tester.pumpWidget(main);
    await tester.tap(find.byType(ElevatedButton));
    await tester.pumpAndSettle();

    verify(methodChannel.invokeMethod("reportErrorStacktrace", any)).called(1);
  });

  testWidgets('Test reportErrorStacktrace', (WidgetTester tester) async {
    Dynatrace dynatrace = DynatraceImpl.private(methodChannel);

    await dynatrace.start(const TestMainWidget());

    Widget main = MaterialApp(
      home: ElevatedButton(
        child: const Text('Test Error'),
        onPressed: () {
          dynatrace.reportErrorStacktrace(
              "FormatException", "ErrorValue", "Invalid Double", "Stacktrace");
        },
      ),
    );

    await tester.pumpWidget(main);
    await tester.tap(find.byType(ElevatedButton));
    await tester.pumpAndSettle();

    verify(methodChannel.invokeMethod("reportErrorStacktrace", any)).called(1);
  });

  testWidgets('Test reportCrashWithException', (WidgetTester tester) async {
    Dynatrace dynatrace = DynatraceImpl.private(methodChannel);

    await dynatrace.start(const TestMainWidget());

    FormatException testException = const FormatException("Invalid Double");

    Widget main = MaterialApp(
      home: ElevatedButton(
        child: const Text('Test Error'),
        onPressed: () {
          dynatrace.reportCrashWithException("FormatException", testException);
        },
      ),
    );

    await tester.pumpWidget(main);
    await tester.tap(find.byType(ElevatedButton));
    await tester.pumpAndSettle();

    verify(methodChannel.invokeMethod("reportCrashWithException", any))
        .called(1);
  });

  testWidgets('Test reportCrash with API', (WidgetTester tester) async {
    Dynatrace dynatrace = DynatraceImpl.private(methodChannel);

    await dynatrace.start(const TestMainWidget());

    Widget main = MaterialApp(
      home: ElevatedButton(
        child: const Text('Test Error'),
        onPressed: () {
          dynatrace.reportCrash(
              "FormatException", "Invalid Double", "Stacktrace");
        },
      ),
    );

    await tester.pumpWidget(main);
    await tester.tap(find.byType(ElevatedButton));
    await tester.pumpAndSettle();

    verify(methodChannel.invokeMethod("reportCrash", any)).called(1);
  });

  test('RunZoned Error', () async {
    when(methodChannel.invokeMethod("reportErrorStacktrace"))
        .thenAnswer((_) => Future<void>.value());

    Dynatrace dynatrace =
        DynatraceImpl.private(methodChannel, startUpFunction: fakeRunApp);

    // Without await here on purpose
    // https://stackoverflow.com/questions/58335158/darts-runzoned-behaviour-on-async-modified-bodies
    dynatrace.start(const TestMainWidget(),
        configuration: const Configuration());

    // Had to insert a delay because the crash handler was not called directly afterwards
    await Future.delayed(const Duration(seconds: 1), () {});

    verify(methodChannel.invokeMethod("reportErrorStacktrace", any)).called(1);
  });

  /*
  IGNORED BECAUSE: https://github.com/flutter/flutter/issues/34499

  testWidgets('Test async error reporting', (WidgetTester tester) async {
    Dynatrace dynatrace = DynatraceImpl.private(methodChannel);

    dynatrace.start(TestMainWidget());

    Widget main = MaterialApp(
      home: RaisedButton(
        child: Text('Test Error'),
        onPressed: () async {
          foo() async {
            try {
              throw FormatException("Invalid Double");
            } catch (e) {
              print("Hell");
            }
          }

          bar() async {
            await foo();
          }

          await bar();
        },
      ),
    );

    await tester.pumpWidget(main);
    await tester.tap(find.byType(RaisedButton));
    await tester.pumpAndSettle();

    verify(methodChannel.invokeMethod("reportErrorStacktrace", any))
        .called(1);
  });

  testWidgets('Test java error reporting', (WidgetTester tester) async {
    Dynatrace dynatrace = DynatraceImpl.private(methodChannel);

    dynatrace.start(TestMainWidget());

    Widget main = MaterialApp(
      home: RaisedButton(
        child: Text('Test Error'),
        onPressed: () async {
          final channel = const MethodChannel('crashy-custom-channel');
          await channel.invokeMethod('blah');
        },
      ),
    );

    await tester.pumpWidget(main);
    await tester.tap(find.byType(RaisedButton));
    await tester.pumpAndSettle();

    verify(methodChannel.invokeMethod("reportErrorStacktrace", any))
        .called(1);
  });*/

// --------- Tests for startWithoutWidget ----------

  testWidgets('Test error reporting using startWithoutWidget',
      (WidgetTester tester) async {
    Dynatrace dynatrace = DynatraceImpl.private(methodChannel);

    await dynatrace.startWithoutWidget();

    Widget main = MaterialApp(
      home: ElevatedButton(
        child: const Text('Test Error'),
        onPressed: () {
          throw const FormatException("Invalid Double");
        },
      ),
    );

    await tester.pumpWidget(main);
    await tester.tap(find.byType(ElevatedButton));
    await tester.pumpAndSettle();

    verify(methodChannel.invokeMethod("reportErrorStacktrace", any)).called(1);
  });

  // Ignoring as mentioned above:
  // testWidgets('Test error reporting async - startWithoutWidget',
  //     (WidgetTester tester) async {
  //   Dynatrace dynatrace = DynatraceImpl.private(methodChannel);

  //   await dynatrace.startWithoutWidget();

  //   Widget main = MaterialApp(
  //     home: ElevatedButton(
  //       child: Text('Test Error'),
  //       onPressed: () async {
  //         throw FormatException("Invalid Double");
  //       },
  //     ),
  //   );

  //   await tester.pumpWidget(main);
  //   await tester.tap(find.byType(ElevatedButton));
  //   await tester.pumpAndSettle();

  //   verify(methodChannel.invokeMethod("reportErrorStacktrace", any)).called(0);
  // });

  test(
      'Test error reporting using startWithoutWidget in zones without reportZoneStacktrace',
      () async {
    when(methodChannel.invokeMethod("reportErrorStacktrace"))
        .thenAnswer((_) => Future<void>.value());

    Dynatrace dynatrace =
        DynatraceImpl.private(methodChannel, startUpFunction: fakeRunApp);

    // Without await here on purpose
    // https://stackoverflow.com/questions/58335158/darts-runzoned-behaviour-on-async-modified-bodies
    runZonedGuarded<Future<void>>(() async {
      dynatrace.startWithoutWidget(configuration: const Configuration());
      fakeRunApp(const TestMainWidget());
      // ignore: avoid_print
    }, (e, s) => print(e));

    // Had to insert a delay because the crash handler was not called directly afterwards
    await Future.delayed(const Duration(seconds: 1), () {});

    verifyNever(methodChannel.invokeMethod("reportErrorStacktrace", any));
  });

  test(
      'Test error reporting using startWithoutWidget with reportZoneStacktrace',
      () async {
    when(methodChannel.invokeMethod("reportErrorStacktrace"))
        .thenAnswer((_) => Future<void>.value());

    Dynatrace dynatrace =
        DynatraceImpl.private(methodChannel, startUpFunction: fakeRunApp);

    // Without await here on purpose
    // https://stackoverflow.com/questions/58335158/darts-runzoned-behaviour-on-async-modified-bodies
    runZonedGuarded<Future<void>>(() async {
      dynatrace.startWithoutWidget(configuration: const Configuration());
      fakeRunApp(const TestMainWidget());
    }, dynatrace.reportZoneStacktrace);

    // Had to insert a delay because the crash handler was not called directly afterwards
    await Future.delayed(const Duration(seconds: 1), () {});

    verify(methodChannel.invokeMethod("reportErrorStacktrace", any)).called(1);
  });
}
