library http_test;

import 'dart:async';

import 'package:dynatrace_flutter_plugin/dynatrace_flutter_plugin.dart';
import 'package:dynatrace_flutter_plugin/src/agent/action/web_request_timing.dart';
import 'package:dynatrace_flutter_plugin/src/agent/agent_impl.dart';
import 'package:dynatrace_flutter_plugin/src/agent/http/http_client.dart';
import 'package:dynatrace_flutter_plugin/src/agent/interface/web_request_timing.dart';
import 'package:flutter/services.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:http/testing.dart';
import 'package:mockito/mockito.dart';
import 'package:http/http.dart';

class _MockDynatrace extends Mock implements DynatraceImpl {
  @override
  Future<String?> getHTTPTagForWebRequest(String? url) {
    return super.noSuchMethod(
        Invocation.method(#getHTTPTagForWebRequest, [url]),
        returnValue: Future.value("TEST_TAG"));
  }

  @override
  WebRequestTiming createWebRequestTiming(String? header, String? url) {
    return super.noSuchMethod(
        Invocation.method(#createWebRequestTiming, [header, url]),
        returnValue:
            DynatraceWebRequestTiming("", "", const MethodChannel("")));
  }

  @override
  DynatraceHttpClient createHttpClient({Client? client}) {
    return super.noSuchMethod(
        Invocation.method(#createHttpClient, [this, client]),
        returnValue: DynatraceHttpClient(this, client: client));
  }
}

class _MockWebRequestTiming extends Mock implements WebRequestTiming {
  @override
  void stopWebRequestTiming(int? responseCode, String? responseMessage) {
    // ignore: void_checks
    return super.noSuchMethod(Invocation.method(
        #stopWebRequestTiming, [responseCode, responseMessage]));
  }
}

void main() {
  late _MockDynatrace dynatrace;

  setUp(() {
    // Set the Dynatrace Singleton to our Mock
    dynatrace = _MockDynatrace();
  });

  test(
      'No header returned with null x-dynatrace value - (Which should not happen nevertheless)',
      () async {
    when(dynatrace.getHTTPTagForWebRequest(any))
        .thenAnswer((_) => Future.value(null));

    var mockClient =
        MockClient((_) async => Response('Testing!', 400, headers: _.headers));

    DynatraceHttpClient dynaClient =
        DynatraceHttpClient(dynatrace, client: mockClient);
    when(dynatrace.createHttpClient(client: any))
        .thenAnswer((realInvocation) => dynaClient);

    var client = dynatrace.createHttpClient(client: mockClient);

    var dynatraceRequest =
        await client.get(Uri.parse('http://some-test-domain.com'));

    expect(dynatraceRequest.statusCode, 400);
    expect(dynatraceRequest.headers["x-dynatrace"], null);
    verifyNever(dynatrace.createWebRequestTiming(any, any));
  });

  test(
      'No header returned with empty x-dynatrace value - (Which should not happen nevertheless)',
      () async {
    when(dynatrace.getHTTPTagForWebRequest(any))
        .thenAnswer((_) => Future.value(""));

    var mockClient =
        MockClient((_) async => Response('Testing!', 400, headers: _.headers));

    DynatraceHttpClient dynaClient =
        DynatraceHttpClient(dynatrace, client: mockClient);
    when(dynatrace.createHttpClient(client: any))
        .thenAnswer((realInvocation) => dynaClient);

    var client = dynatrace.createHttpClient(client: mockClient);

    var dynatraceRequest =
        await client.get(Uri.parse('http://some-test-domain.com'));

    expect(dynatraceRequest.statusCode, 400);
    expect(dynatraceRequest.headers["x-dynatrace"], null);
    verifyNever(dynatrace.createWebRequestTiming(any, any));
  });

  test('Request Working', () async {
    when(dynatrace.getHTTPTagForWebRequest(any))
        .thenAnswer((_) => Future.value("MT_Tag"));

    var mockClient =
        MockClient((_) async => Response('Testing!', 200, headers: _.headers));

    _MockWebRequestTiming timing = _MockWebRequestTiming();

    when(dynatrace.createWebRequestTiming(any, any))
        .thenAnswer((realInvocation) => timing);
    DynatraceHttpClient dynaClient =
        DynatraceHttpClient(dynatrace, client: mockClient);
    when(dynatrace.createHttpClient(client: any))
        .thenAnswer((realInvocation) => dynaClient);

    var client = dynatrace.createHttpClient(client: mockClient);

    var dynatraceRequest =
        await client.get(Uri.parse('http://some-test-domain.com'));

    expect(client.runtimeType, DynatraceHttpClient);
    expect(dynatraceRequest.statusCode, 200);
    expect(dynatraceRequest.headers["x-dynatrace"], "MT_Tag");
    verify(timing.startWebRequestTiming()).called(1);
    verify(timing.stopWebRequestTiming(any, any)).called(1);
  });

  test('Request throwing an error', () async {
    when(dynatrace.getHTTPTagForWebRequest(any))
        .thenAnswer((_) => Future.value("MT_Tag"));

    var mockClient = MockClient((_) async {
      Response('Testing!', 503, headers: _.headers);
      throw StateError("ouch!");
    });

    _MockWebRequestTiming timing = _MockWebRequestTiming();

    when(dynatrace.createWebRequestTiming(any, any))
        .thenAnswer((realInvocation) => timing);
    DynatraceHttpClient dynaClient =
        DynatraceHttpClient(dynatrace, client: mockClient);
    when(dynatrace.createHttpClient(client: any))
        .thenAnswer((realInvocation) => dynaClient);

    var client = dynatrace.createHttpClient(client: mockClient);

    expect(client.runtimeType, DynatraceHttpClient);
    try {
      await client.get(Uri.parse('http://some-test-domain.com'));
    } catch (e) {
      verify(timing.startWebRequestTiming()).called(1);
      verify(timing.stopWebRequestTiming(-1, any)).called(1);
    }
  });
}
