import 'package:dynatrace_flutter_plugin/src/cli/config/model/android_config.dart';
import 'package:dynatrace_flutter_plugin/src/cli/config/android_configurator.dart';
import 'package:dynatrace_flutter_plugin/src/cli/config/model/cli_arguments.dart';
import 'package:dynatrace_flutter_plugin/src/cli/util/pathConstants.dart';

import 'package:flutter_test/flutter_test.dart';
import 'package:mockito/mockito.dart';
import 'package:path/path.dart' as p;
import 'dart:io';

String currentFolder =
    p.current.endsWith("test") ? p.join(p.current, "..") : p.current;
String gradleOrigPath = p.normalize(
    p.join(currentFolder, "test", "cli", "config", "gradle", "build.gradle"));
String gradleCopyPath = p.normalize(
    p.join(currentFolder, "test", "cli", "config", "build.copy.gradle"));
String gradleDynatracePath = p.normalize(
    p.join(currentFolder, "test", "cli", "config", "dynatrace.gradle"));
String gradlePluginPath = p
    .normalize(p.join(currentFolder, "test", "cli", "config", "plugin.gradle"));

String dynatraceGradle = 'apply from: "./dynatrace.gradle"';
String pluginGradle = 'apply from: "./plugin.gradle", to: buildscript';

class MockPaths extends Mock implements PathConstants {
  @override
  String getDynatraceGradle(Directory? directory) {
    return super.noSuchMethod(
        Invocation.method(#getDynatraceGradle, [directory]),
        returnValue: "");
  }

  @override
  String getPluginGradle(Directory? gradleDir) {
    return super.noSuchMethod(Invocation.method(#getPluginGradle, [gradleDir]),
        returnValue: "");
  }
}

class MockCLI extends Mock implements CommandLineArguments {}

void main() {
  setUp(() {
    return gradleSetup();
  });

  tearDown(() {
    return gradleTearDown();
  });

  test('Modify Root Build.Gradle', () async {
    File gradleFile = File(gradleCopyPath);

    String contentGradleFile = await gradleFile.readAsString();
    expect(contentGradleFile.contains(dynatraceGradle), false);
    expect(contentGradleFile.contains(pluginGradle), false);

    AndroidConfigurator configurator =
        AndroidConfigurator(MockPaths(), MockCLI(), File(gradleCopyPath));

    await configurator.instrumentAndroidPlatform();

    contentGradleFile = await gradleFile.readAsString();
    expect(contentGradleFile.contains(dynatraceGradle), true);
    expect(contentGradleFile.contains(pluginGradle), true);
  });

  test('File not available', () async {
    AndroidConfigurator configurator = AndroidConfigurator(
        MockPaths(), MockCLI(), File("FileNotAvailable.gradle"));

    try {
      await configurator.instrumentAndroidPlatform();
      fail("No Exception thrown!");
    } catch (error) {
      expect(
          error.toString().contains(
              "Can't find .gradle file. Gradle path must also include the gradle file!"),
          true);
      expect(error.runtimeType, FileSystemException);
    }
  });

  test('Other File but not gradle', () async {
    AndroidConfigurator configurator =
        AndroidConfigurator(MockPaths(), MockCLI(), File("SomeJavaFile.java"));

    try {
      await configurator.instrumentAndroidPlatform();
      fail("No Exception thrown!");
    } catch (error) {
      expect(
          error.toString().contains(
              "Can't find .gradle file. Gradle path must also include the gradle file!"),
          true);
      expect(error.runtimeType, FileSystemException);
    }
  });

  test('No Configuration available', () async {
    AndroidConfiguration config = AndroidConfiguration(null);

    AndroidConfigurator configurator =
        AndroidConfigurator(MockPaths(), MockCLI(), File(gradleCopyPath));

    await configurator.writeGradleConfig(config);

    expect(await File(gradlePluginPath).exists(), isFalse);
    expect(await File(gradleDynatracePath).exists(), isFalse);
  });

  test('Configuration available', () async {
    AndroidConfiguration config = AndroidConfiguration("TestConfiguration");

    MockPaths paths = MockPaths();
    when(paths.getDynatraceGradle(any)).thenReturn(gradleDynatracePath);
    when(paths.getPluginGradle(any)).thenReturn(gradlePluginPath);
    AndroidConfigurator configurator =
        AndroidConfigurator(paths, MockCLI(), File(gradleCopyPath));

    await configurator.writeGradleConfig(config);

    expect(await File(gradlePluginPath).exists(), isTrue);
    expect(await File(gradleDynatracePath).exists(), isTrue);

    String fileContent = await File(gradleDynatracePath).readAsString();
    expect(fileContent.contains("TestConfiguration"), isTrue);
  });

  test('Uninstall', () async {
    // Add => Remove gradle files and build.gradle configs in build.gradle
    File gradleFile = File(gradleCopyPath);
    AndroidConfiguration config = AndroidConfiguration("UninstallTest");

    MockPaths paths = MockPaths();
    when(paths.getDynatraceGradle(any)).thenReturn(gradleDynatracePath);
    when(paths.getPluginGradle(any)).thenReturn(gradlePluginPath);
    AndroidConfigurator configurator =
        AndroidConfigurator(paths, MockCLI(), File(gradleCopyPath));

    String contentGradleFile = await gradleFile.readAsString();
    expect(contentGradleFile.contains(dynatraceGradle), false);
    expect(contentGradleFile.contains(pluginGradle), false);

    // Add apply dynatrace/plugin.gradle lines to build.gradle file
    await configurator.instrumentAndroidPlatform();

    contentGradleFile = await gradleFile.readAsString();
    expect(contentGradleFile.contains(dynatraceGradle), true);
    expect(contentGradleFile.contains(pluginGradle), true);

    // Copy dynatrace/plugin.gradle files
    await configurator.writeGradleConfig(config);

    expect(await File(gradlePluginPath).exists(), isTrue);
    expect(await File(gradleDynatracePath).exists(), isTrue);

    // Remove apply dynatrace/plugin.gradle lines from build.gradle file
    await configurator.removeGradleLinesForUninstall();

    contentGradleFile = await gradleFile.readAsString();
    expect(contentGradleFile.contains(dynatraceGradle), false);
    expect(contentGradleFile.contains(pluginGradle), false);

    // Remove dynatrace.gradle and plugin.gradle files
    await configurator.removeDynatraceGradleFiles();

    expect(await File(gradlePluginPath).exists(), isFalse);
    expect(await File(gradleDynatracePath).exists(), isFalse);
  });
}

/// Makes a Copy of the gradle file so we can work with it
Future<void> gradleSetup() async {
  File gradleFile = File(gradleOrigPath);
  await gradleFile.copy(gradleCopyPath);
}

Future<void> gradleTearDown() async {
  File gradleCopyFile = File(gradleCopyPath);
  await gradleCopyFile.delete();

  File gradleDynatraceFile = File(gradleDynatracePath);
  if (await gradleDynatraceFile.exists()) {
    await gradleDynatraceFile.delete();
  }

  File gradlePluginFile = File(gradlePluginPath);
  if (await gradlePluginFile.exists()) {
    await gradlePluginFile.delete();
  }
}
