import 'package:dynatrace_flutter_plugin/src/agent/util/logger.dart';
import 'package:dynatrace_flutter_plugin/src/cli/config/model/cli_arguments.dart';
import 'package:dynatrace_flutter_plugin/src/cli/config/model/ios_config.dart';
import 'package:dynatrace_flutter_plugin/src/cli/config/ios_configurator.dart';
import 'package:dynatrace_flutter_plugin/src/cli/util/pathConstants.dart';
import 'package:dynatrace_flutter_plugin/src/cli/util/util.dart';

import 'package:flutter_test/flutter_test.dart';
import 'package:mockito/mockito.dart';
import 'package:path/path.dart' as p;
import 'dart:io';

String currentFolder =
    p.current.endsWith("test") ? p.join(p.current, "..") : p.current;
String pListOrigPath = p.normalize(
    p.join(currentFolder, "test", "cli", "config", "plist", "Info.plist"));
String pListCopyPath = p.normalize(
    p.join(currentFolder, "test", "cli", "config", "Info.copy.plist"));
String dynatracePlistPath = p.normalize(
    p.join(currentFolder, "test", "cli", "config", "Dynatrace.plist"));

class MockCLI extends Mock implements CommandLineArguments {}

class MockLogger extends Mock implements Logger {}

class MockUtil extends Mock implements Util {
  @override
  String currentOS() {
    return super.noSuchMethod(Invocation.method(#currentOS, []),
        returnValue: Util().currentOS());
  }

  @override
  bool isMacOS() {
    return super.noSuchMethod(Invocation.method(#isMacOS, []),
        returnValue: Util().isMacOS());
  }
}

class MockPaths extends Mock implements PathConstants {
  @override
  String getIosDynatracePListFile() {
    return super.noSuchMethod(Invocation.method(#getIosDynatracePListFile, []),
        returnValue: "");
  }
}

void main() {
  MockUtil util = MockUtil();
  MockLogger logger = MockLogger();

  /// Makes a Copy of the PList File so we can work with it
  Future<void> pListSetup() async {
    File pListFile = File(pListOrigPath);
    await pListFile.copy(pListCopyPath);
  }

  Future<void> pListTearDown() async {
    File pListCopyFile = File(pListCopyPath);
    await pListCopyFile.delete();
    File dynatracePlistFile = File(dynatracePlistPath);
    if (await dynatracePlistFile.exists()) {
      await dynatracePlistFile.delete();
    }
  }

  setUp(() {
    return pListSetup();
  });

  tearDown(() {
    return pListTearDown();
  });

  group('iOS Config Tests - MacOS - ', () {
    setUp(() {
      when(util.currentOS()).thenReturn("macos");
      when(util.isMacOS()).thenReturn(true);
    });

    tearDown(() {
      resetMockitoState();
    });

    test('No Configuration', () async {
      File pListFile = File(pListCopyPath);
      String fileContent = await pListFile.readAsString();

      IosConfigurator configurator =
          IosConfigurator(MockCLI(), File(pListCopyPath), util, logger);
      await configurator.modifyPListFile(IosConfiguration(null));

      String fileContentAfterwards = await pListFile.readAsString();

      // PList Content Stays untouched
      expect(fileContent.replaceAll("\r\n", "\n"), fileContentAfterwards);

      // Pre Fill PList File and see that it removes the values
      String dynatraceConfigSnippet = "<key>DTXWhatEver</key>\n<true/>";
      IosConfiguration config = IosConfiguration(dynatraceConfigSnippet);
      await configurator.modifyPListFile(config);

      fileContentAfterwards = await pListFile.readAsString();
      expect(fileContentAfterwards.contains("DTXWhatEver"), isTrue);
      expect(fileContentAfterwards.contains("DTXFlavor"), isTrue);

      // Modify again without configuration not it should keep the original
      // config unless a new config is being added.
      await configurator.modifyPListFile(IosConfiguration(null));

      verify(
          logger.d('Current operating system: macos', logType: LogType.Info));

      fileContentAfterwards = await pListFile.readAsString();
      expect(fileContentAfterwards.contains("DTXWhatEver"), isTrue);
      expect(fileContentAfterwards.contains("DTXFlavor"), isTrue);
    });

    test('Empty Configuration', () async {
      File pListFile = File(pListCopyPath);
      String fileContent = await pListFile.readAsString();

      IosConfigurator configurator =
          IosConfigurator(MockCLI(), File(pListCopyPath), util, logger);
      IosConfiguration config = IosConfiguration("");
      await configurator.modifyPListFile(config);

      verify(
          logger.d('Current operating system: macos', logType: LogType.Info));

      String fileContentAfterwards = await pListFile.readAsString();

      // PList Content Stays untouched
      expect(fileContent.replaceAll("\r\n", "\n"), fileContentAfterwards);
    });

    test('No prefilled Configuration', () async {
      File pListFile = File(pListCopyPath);

      String dynatraceConfigSnippet = "<key>DTXWhatEver</key>\n<true/>";
      IosConfiguration config = IosConfiguration(dynatraceConfigSnippet);

      IosConfigurator configurator =
          IosConfigurator(MockCLI(), File(pListCopyPath), util, logger);
      await configurator.modifyPListFile(config);

      verify(
          logger.d('Current operating system: macos', logType: LogType.Info));

      String fileContentAfterwards = await pListFile.readAsString();
      expect(fileContentAfterwards.contains("DTXWhatEver"), isTrue);
      expect(fileContentAfterwards.contains("DTXFlavor"), isTrue);
    });

    test('Prefilled Configuration', () async {
      MockPaths paths = MockPaths();
      when(paths.getIosDynatracePListFile()).thenReturn(dynatracePlistPath);
      File pListFile = File(pListCopyPath);

      String dynatraceConfigSnippet = "<key>DTXWhatEver</key>\n<true/>";
      IosConfiguration config = IosConfiguration(dynatraceConfigSnippet);

      IosConfigurator configurator =
          IosConfigurator(MockCLI(), File(pListCopyPath), util, logger);
      await configurator.modifyPListFile(config);

      dynatraceConfigSnippet = "<key>DTXWhatOther</key>\n<true/>";
      config = IosConfiguration(dynatraceConfigSnippet);

      await configurator.modifyPListFile(config);

      verify(
          logger.d('Current operating system: macos', logType: LogType.Info));

      String fileContentAfterwards = await pListFile.readAsString();

      // ignore: avoid_print
      print(fileContentAfterwards);

      expect(fileContentAfterwards.contains("DTXWhatOther"), isTrue);
      expect(fileContentAfterwards.contains("DTXWhatEver"), isFalse);
      expect(fileContentAfterwards.contains("DTXFlavor"), isTrue);
    });

    test('File not available', () async {
      IosConfigurator configurator =
          IosConfigurator(MockCLI(), File("FileNotFound.plist"), util, logger);

      expect(
          () => configurator.modifyPListFile(IosConfiguration(null)),
          throwsA(predicate((dynamic e) =>
              e is FileSystemException &&
              e.message ==
                  'Can\'t find .plist file. plist path must also include the plist file! : FileNotFound.plist')));
      verify(
          logger.d('Current operating system: macos', logType: LogType.Info));
    });

    test('File not ending with .plist', () async {
      IosConfigurator configurator = IosConfigurator(
          MockCLI(), File("FileWithDifferentExtension.java"), util, logger);

      expect(
          () => configurator.modifyPListFile(IosConfiguration(null)),
          throwsA(predicate((dynamic e) =>
              e is FileSystemException &&
              e.message ==
                  'Can\'t find .plist file. plist path must also include the plist file! : FileWithDifferentExtension.java')));

      verify(
          logger.d('Current operating system: macos', logType: LogType.Info));
    });

    test('Uninstall - Remove properties from Plist', () async {
      File pListFile = File(pListCopyPath);

      String dynatraceConfigSnippet = "<key>DTXUninstall</key>\n<true/>";
      IosConfiguration config = IosConfiguration(dynatraceConfigSnippet);

      IosConfigurator configurator =
          IosConfigurator(MockCLI(), File(pListCopyPath), util, logger);
      await configurator.modifyPListFile(config);

      String fileContentAfterwards = await pListFile.readAsString();
      expect(fileContentAfterwards.contains("DTXUninstall"), isTrue);
      expect(fileContentAfterwards.contains("DTXFlavor"), isTrue);

      // Mimic uninstall by removing Dynatrace properties in plist
      await configurator.modifyPListFile(config, isUninstall: true);
      fileContentAfterwards = await pListFile.readAsString();

      expect(fileContentAfterwards.contains("DTXUninstall"), isTrue);
      expect(fileContentAfterwards.contains("DTXUninstall"), isTrue);
      verify(
          logger.d('Current operating system: macos', logType: LogType.Info));
      verify(logger.d('Removing Dynatrace properties from $pListCopyPath!',
          logType: LogType.Info));
    });
  });

  group('iOS Config Tests - Non MacOS - ', () {
    setUp(() {
      when(util.currentOS()).thenReturn("windows");
      when(util.isMacOS()).thenReturn(false);
    });

    tearDown(() {
      resetMockitoState();
    });

    test('No Configuration', () async {
      File pListFile = File(pListCopyPath);
      String fileContent = await pListFile.readAsString();

      IosConfigurator configurator =
          IosConfigurator(MockCLI(), File(pListCopyPath), util, logger);
      await configurator.modifyPListFile(IosConfiguration(null));
      expect(util.currentOS(), "windows");

      String fileContentAfterwards = await pListFile.readAsString();

      // PList Content Stays untouched
      expect(fileContent.replaceAll("\r\n", "\n"), fileContentAfterwards);

      // Pre Fill PList File and see that it removes the values
      String dynatraceConfigSnippet = "<key>DTXWhatEver</key>\n<true/>";
      IosConfiguration config = IosConfiguration(dynatraceConfigSnippet);
      await configurator.modifyPListFile(config);

      verify(
          logger.d('Current operating system: windows', logType: LogType.Info));
      verify(logger.i(
          'Can\'t write configuration of iOS agent as current operating system is not MacOS!',
          logType: LogType.Warning));

      fileContentAfterwards = await pListFile.readAsString();
      expect(!fileContentAfterwards.contains("DTXWhatEver"), isTrue);
      expect(!fileContentAfterwards.contains("DTXFlavor"), isTrue);
    });

    test('Empty Configuration', () async {
      File pListFile = File(pListCopyPath);
      String fileContent = await pListFile.readAsString();

      IosConfigurator configurator =
          IosConfigurator(MockCLI(), File(pListCopyPath), util, logger);
      IosConfiguration config = IosConfiguration("");
      await configurator.modifyPListFile(config);

      verify(
          logger.d('Current operating system: windows', logType: LogType.Info));
      verify(logger.i(
          'Can\'t write configuration of iOS agent as current operating system is not MacOS!',
          logType: LogType.Warning));

      String fileContentAfterwards = await pListFile.readAsString();

      // PList Content Stays untouched
      expect(fileContent.replaceAll("\r\n", "\n"), fileContentAfterwards);
    });

    test('No prefilled Configuration', () async {
      File pListFile = File(pListCopyPath);

      String dynatraceConfigSnippet = "<key>DTXWhatEver</key>\n<true/>";
      IosConfiguration config = IosConfiguration(dynatraceConfigSnippet);

      IosConfigurator configurator =
          IosConfigurator(MockCLI(), File(pListCopyPath), util, logger);
      await configurator.modifyPListFile(config);

      verify(
          logger.d('Current operating system: windows', logType: LogType.Info));
      verify(logger.i(
          'Can\'t write configuration of iOS agent as current operating system is not MacOS!',
          logType: LogType.Warning));

      String fileContentAfterwards = await pListFile.readAsString();
      expect(fileContentAfterwards.contains("DTXWhatEver"), isFalse);
      expect(fileContentAfterwards.contains("DTXFlavor"), isFalse);
    });

    test('Prefilled Configuration', () async {
      MockPaths paths = MockPaths();
      when(paths.getIosDynatracePListFile()).thenReturn(dynatracePlistPath);

      String dynatraceConfigSnippet = "<key>DTXWhatEver</key>\n<true/>";
      IosConfiguration config = IosConfiguration(dynatraceConfigSnippet);

      IosConfigurator tempConfigurator =
          IosConfigurator(MockCLI(), File(pListCopyPath), util, logger);
      File testPlist = tempConfigurator.createTempPlist(config);

      IosConfigurator configurator =
          IosConfigurator(MockCLI(), File(dynatracePlistPath), util, logger);

      dynatraceConfigSnippet = "<key>DTXWhatOther</key>\n<true/>";
      config = IosConfiguration(dynatraceConfigSnippet);

      await configurator.modifyPListFile(config);

      verify(
          logger.d('Current operating system: windows', logType: LogType.Info));
      verify(logger.i(
          'Can\'t write configuration of iOS agent as current operating system is not MacOS!',
          logType: LogType.Warning));

      String fileContentAfterwards = await testPlist.readAsString();

      // ignore: avoid_print
      print(fileContentAfterwards);

      expect(fileContentAfterwards.contains("DTXWhatOther"), isFalse);
      expect(fileContentAfterwards.contains("DTXWhatEver"), isTrue);
      expect(fileContentAfterwards.contains("DTXFlavor"), isTrue);
    });

    test('File not available', () async {
      IosConfigurator configurator =
          IosConfigurator(MockCLI(), File("FileNotFound.plist"), util, logger);

      configurator.modifyPListFile(IosConfiguration(null));

      verify(
          logger.d('Current operating system: windows', logType: LogType.Info));
      verify(logger.i(
          'Can\'t write configuration of iOS agent as current operating system is not MacOS!',
          logType: LogType.Warning));
    });

    test('File not ending with .plist', () async {
      IosConfigurator configurator = IosConfigurator(
          MockCLI(), File("FileWithDifferentExtension.java"), util, logger);

      configurator.modifyPListFile(IosConfiguration(null));

      verify(
          logger.d('Current operating system: windows', logType: LogType.Info));
      verify(logger.i(
          'Can\'t write configuration of iOS agent as current operating system is not MacOS!',
          logType: LogType.Warning));
    });

    test('Uninstall - Remove properties from Plist', () async {
      String dynatraceConfigSnippet = "<key>DTXUninstall</key>\n<true/>";
      IosConfiguration config = IosConfiguration(dynatraceConfigSnippet);

      IosConfigurator configurator =
          IosConfigurator(MockCLI(), File(pListCopyPath), util, logger);
      await configurator.modifyPListFile(config, isUninstall: true);

      verify(
          logger.d('Current operating system: windows', logType: LogType.Info));
      verify(logger.i(
          "Unable to remove Dynatrace properties from $pListCopyPath as current operating system is not MacOS!",
          logType: LogType.Warning));
    });
  });
}
